<div v-if="data.{{{widget.widget_id}}}.widget_routes && data.{{{widget.widget_id}}}.widget_routes.cell_clicked_cb">
    <sly-table :content="data.{{{widget.widget_id}}}.table_data"
               :options="data.{{{widget.widget_id}}}.table_options"
               @input="state.{{{widget.widget_id}}}.selected_row = $event; post('/{{{widget.widget_id}}}/cell_clicked_cb')">

    </sly-table>
</div>

<div v-if="!data.{{{widget.widget_id}}}.widget_routes || !data.{{{widget.widget_id}}}.widget_routes.cell_clicked_cb">
    <sly-table :content="data.{{{widget.widget_id}}}.table_data"
               :options="data.{{{widget.widget_id}}}.table_options"
    >
    </sly-table>
</div>
